#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "lnm/http/loop.h"
#include "lnm/http/route.h"
#include "lnm/log.h"

#include "lander.h"

const char *lander_server = "lander/" LANDER_VERSION;

lnm_http_loop *loop_init(lander_gctx *gctx, const char *api_key) {
  lnm_http_loop *hl;
  lnm_http_loop_init(&hl, gctx, lander_ctx_init,
                     (lnm_http_ctx_reset_fn)lander_ctx_reset,
                     (lnm_http_ctx_free_fn)lander_ctx_free);
  lnm_http_loop_set_api_key(hl, api_key);
  lnm_http_loop_set_server(hl, lander_server);

  lnm_http_router *router;
  lnm_http_router_init(&router);

  lnm_http_route *route;
  lnm_http_router_add(&route, router, lnm_http_method_get, "/");
  lnm_http_route_step_append(route, lander_get_index, false);

  lnm_http_router_add(&route, router, lnm_http_method_get, "/:key");
  lnm_http_route_step_append(route, lander_get_entry, false);

  lnm_http_router_add(&route, router, lnm_http_method_delete, "/:key");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_remove_entry, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/s/");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_redirect, false);
  lnm_http_route_step_append(route, lnm_http_loop_step_body_to_buf, false);
  lnm_http_route_step_append(route, lander_post_redirect_body_to_attr, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/sl/");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_redirect_secure, false);
  lnm_http_route_step_append(route, lnm_http_loop_step_body_to_buf, false);
  lnm_http_route_step_append(route, lander_post_redirect_body_to_attr, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/s/:key");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_redirect, false);
  lnm_http_route_step_append(route, lnm_http_loop_step_body_to_buf, false);
  lnm_http_route_step_append(route, lander_post_redirect_body_to_attr, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/p/");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_paste, false);
  lnm_http_route_step_append(route, lander_stream_body_to_entry, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/pl/");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_paste_secure, false);
  lnm_http_route_step_append(route, lander_stream_body_to_entry, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/p/:key");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_paste, false);
  lnm_http_route_step_append(route, lander_stream_body_to_entry, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/f/");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_file, false);
  lnm_http_route_step_append(route, lander_stream_body_to_entry, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/fl/");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_file_secure, false);
  lnm_http_route_step_append(route, lander_stream_body_to_entry, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_router_add(&route, router, lnm_http_method_post, "/f/:key");
  lnm_http_route_step_append(route, lnm_http_loop_step_auth, false);
  lnm_http_route_step_append(route, lander_post_file, false);
  lnm_http_route_step_append(route, lander_stream_body_to_entry, false);
  lnm_http_route_step_append(route, lander_commit_entry, true);

  lnm_http_loop_router_set(hl, router);

  return hl;
}

#define ENV(var, env_var)                                                      \
  const char *var = getenv(env_var);                                           \
  if (var == NULL) {                                                           \
    lnm_lcritical("main", "Missing environment variable %s", env_var);         \
    exit(1);                                                                   \
  }

#define ENV_OPT(var, env_var, default)                                         \
  const char *var = getenv(env_var);                                           \
  if (var == NULL) {                                                           \
    var = default;                                                             \
  }

int main() {
  srand(time(NULL));

  lnm_log_init_global();
  lnm_log_register_stdout(lnm_log_level_info);

  ENV(api_key, "LANDER_API_KEY");
  ENV_OPT(port_str, "LANDER_PORT", "18080");
  ENV_OPT(data_dir_s, "LANDER_DATA_DIR", ".");

  int port = atoi(port_str);

  if (port <= 0 || port >= 1 << 16) {
    lnm_lcritical("main", "Invalid TCP port %s", port_str);
    exit(1);
  }

  lander_gctx *c_gctx = lander_gctx_init();
  c_gctx->data_dir = data_dir_s;

  lsm_str *data_dir;
  lsm_str_init_copy(&data_dir, (char *)data_dir_s);

  lnm_linfo("main", "Initializing store from path '%s'", data_dir_s);

  if (lsm_store_load(&c_gctx->store, data_dir) != lsm_error_ok) {
    lnm_lcritical("main", "%s", "Failed to load existing store.");
  }

  lnm_linfo("main", "Store loaded containing %lu entries",
            lsm_store_size(c_gctx->store));
  lnm_http_loop *hl = loop_init(c_gctx, api_key);
  lnm_http_loop_run(hl, port, 1, 0);
}
